local WYM = _G.WymBuffs
local L = WYM.L

local buffFrames = {}
local cooldowns = {}

local dispelTypePriority = { ["Poison"] = 1, ["Magic"] = 2, ["Curse"] = 3, ["Disease"] = 4, ["none"] = 5, ["warning"] = 0, }

function WYM:CreateBuffFrame(id)
    frame = CreateFrame("FRAME", nil, UIParent)
    frame.bars = {}
    frame.icons = {}
    frame.auras = {} -- Unfiltered list of auras of the unit/type for this frame
    frame.buffs = {} -- The filtered list of buffs we want this frame to show
    frame:SetWidth(1)
    frame:SetHeight(1)
    frame:SetPoint("CENTER", UIParent, "CENTER")
    frame:SetAlpha(1)
    buffFrames[id] = frame
    return frame
end

function WYM:DeleteBuffFrame(id)
    self:ResetBuffFrame(buffFrames[id])
    buffFrames[id] = nil
end

function WYM:GetBuffFrame(id)
    return buffFrames[id]
end

function WYM:GetBuffFrames()
    return buffFrames
end

function WYM:UpdateBuffFrames(unit)
    for id, frame in pairs(buffFrames) do
        if not unit or frame.settings.unit == unit then
            self:UpdateBuffFrame(frame)
        end
    end
end

function WYM:UpdateBuffFrame(frame)
    
    self:ResetBuffFrame(frame)
    
    if WYM:IsEnabled() and frame.settings.enabled then

        if frame.settings.hideOutOfCombat and not UnitAffectingCombat("player") then return end

        if frame.settings.showForNPCsOnly and UnitIsPlayer(frame.settings.unit) then return end
        
        if not frame:IsVisible() then
            frame:Show()
        end        
        
        local buffs = frame.buffs
        
        if frame.settings.unit == "weapon" then
            buffs = self:GetTemporaryWeaponEnchants(frame)
        elseif frame.settings.showMissingBuffs then
            buffs = self:GetMissingAuras(frame)
        else
            for _, buff in ipairs(self:GetUnitAuras(frame)) do
                if self:ShowBuff(frame, buff) then
                    tinsert(buffs, buff)
                end
            end
        end
        
        if frame.settings.showTestBuffs then
            for _, aura in ipairs(self:GetDummyAuras()) do
                tinsert(buffs, aura)
            end
        end            

        buffs = self:SortBuffs(frame, buffs, frame.settings.sortOrder, frame.settings.reverseSortOrder)

        local width, height
        if frame.settings.displayType == "BARS" then
            self:UpdateBars(frame, buffs)
            width, height = self:GetBuffBarFrameSize(frame)
        else
            self:UpdateIcons(frame, buffs)
            width, height = self:GetBuffIconFrameSize(frame)
        end

        frame:SetWidth(width)
        frame:SetHeight(height)

        frame:ClearAllPoints()
        frame:SetPoint(frame.settings.positionAnchorPoint,
                       UIParent,
                       frame.settings.positionRelativePoint,
                       frame.settings.positionXOffset,
                       frame.settings.positionYOffset)
    else
        frame:Hide()
    end
end

function WYM:ResetBuffFrame(frame)
    self:ResetBars(frame)
    self:ResetIcons(frame)
    wipe(frame.bars)
    wipe(frame.icons)
    self:RecycleAuras(frame)
    wipe(frame.auras)
    wipe(frame.buffs)    
end

function WYM:SortBuffs(frame, buffs, sortOrder, reverseSort)
    if sortOrder == "NAME" then
        table.sort(buffs, function(a,b) return a.name < b.name end)        
    elseif sortOrder == "DURATION" then
        table.sort(buffs, function(a,b)
                            -- This hack sorts buffs with no duration last instead of first
                            local durationa = a.duration
                            local durationb = b.duration
                            if durationa == 0 then durationb = -1 end
                            if durationb == 0 then durationa = -1 end
                            return durationa < durationb
                          end)
    elseif sortOrder == "TYPE" then
        table.sort(buffs, function(a,b)
                            local atype = a.dispelType
                            if (not atype) or atype == "" then atype = "none" end
                            local btype = b.dispelType
                            if (not btype) or btype == "" then btype = "none" end
                            
                            if dispelTypePriority[atype] == dispelTypePriority[btype] then
                                return a.name < b.name
                            else
                                return dispelTypePriority[atype] < dispelTypePriority[btype]
                            end
                          end)
    elseif sortOrder == "ADDED" then
        -- This is the order we get the buffs from UnitAura() in so no need for extra sorting
    end
    
    if reverseSort then
        i = # buffs
        j = 1
        reverse = {}
        while i > 0 do
            reverse[j] = buffs[i]
            i = i - 1
            j = j + 1
        end
        return reverse
    end
    return buffs
end

function WYM:ShowBuff(frame, buff)
    if buff.isTest then return true end
        
    if frame.settings.enableDebuffCategoryFilter then
        return self:IsEnabledDebuffCategory(frame, buff)    
    end

    local visible = false
  
    if frame.settings.buffWatchList[buff.name] then
        visible = not visible
    end
    
    if frame.settings.blacklist then
        visible = not visible    
    end    
    
    if not visible then return false end

    if frame.settings.showOwnBuffsOnly then
        if not buff.unitCaster or not (buff.unitCaster == "player") then
            return false
        end
    end
        
    if frame.settings.showDispellableOnly and (not buff.dispelType or not buff.dispelType == "Magic") then
        return false
    end
    
    if frame.settings.showStealableOnly and not buff.isStealable then
        return false
    end

    if not frame.settings.showNoDurationBuffs and buff.duration <= 0 then
        return false
    end

    if not frame.settings.showDurationBuffs and buff.duration > 0 then
        return false
    end

    local maxDuration = tonumber(frame.settings.showMaxDuration)
    if maxDuration and maxDuration > 0 and (buff.duration <= 0 or maxDuration < buff.duration) then    
        return false
    end
    
    return true
end

function WYM:EnableWarnings(frame, buff)
    -- There has to be a better way to check if a buff is 'new', someday I might even find it
    if buff.duration - buff.remaining() < 0.1 and not buff.isTemporaryWeaponEnchant then
        if frame.settings.playWarningSoundOnNewBuff or frame.settings.showWarningFlashOnNewBuff then
            buff.newBuffWarning = true
        end            
    end

    local buffSettings = frame.settings.buffWatchList[buff.name]
    if buffSettings then
        if buffSettings.warnWhenExpiring then
            buff.warningSoundTime = buff.expirationTime - buffSettings.expirationWarningTime
            buff.warningBlinkTime = buff.warningSoundTime
            if buff.warningSoundTime <= GetTime() then
                buff.warningSoundTime = nil
            end
        end
    end
end

function WYM:HandleWarnings(frame, buff)    
    if buff.newBuffWarning then        
        buff.newBuffWarning = nil
        
        if frame.settings.playWarningSoundOnNewBuff then
            PlaySoundFile(self:GetWarningSound(frame.settings.warningSoundOnNewBuff))
        end

        if frame.settings.showWarningFlashOnNewBuff then        
            local cooldown = tremove(cooldowns)
            if not cooldown then
                cooldown = CreateFrame("Cooldown",nil,frame)            
            end
    
            cooldown:ClearAllPoints()
            if frame.settings.displayType == "BARS" then
                cooldown:SetPoint("CENTER",buff.widget.candyBarIconFrame,"CENTER")
            else
                cooldown:SetPoint("CENTER",buff.widget,"CENTER")
            end
            cooldown:SetSize(frame.settings.warningFlashOnNewBuffSize,frame.settings.warningFlashOnNewBuffSize)
            cooldown:SetCooldown(0,0)
            tinsert(cooldowns,cooldown)
        end
    end

    local buffSettings = frame.settings.buffWatchList[buff.name]
    
    if buffSettings and (not buffSettings.onlyWarnInCombat or UnitAffectingCombat("player")) then
        if buffSettings.playWarningSound and buff.warningSoundTime and buff.warningSoundTime < GetTime() then
            PlaySoundFile(self:GetWarningSound(buffSettings.warningSound))
            buff.warningSoundTime = nil
        end

        if buffSettings.warningBlinkBuff and buff.warningBlinkTime and buff.warningBlinkTime < GetTime() then
            self:blink(buff.widget, buff.remaining(), buffSettings.warningBlinkBuffSpeed)
        end
    end
end

function WYM:blink(widget, duration, speed)
    -- Apparently there are issues with UIFrameFlash and taint if you use the built
    -- in voicechat so unless Blizzard fixes that we'll eventually have to implement
    -- this some other way
    local s = tonumber(speed)
    UIFrameFlash(widget, s, s, duration, true, s, s)
end

function WYM:stopBlink(widget)
    UIFrameFlashStop(widget)
end
